from enum import Enum


class Platform(str, Enum):
    win = 'win'
    linux = 'linux'
    mac = 'mac'
    win32 = 'win32'
    win64 = 'win64'
    linux32 = 'linux32'
    linux64 = 'linux64'
    mac32 = 'mac32'
    mac64 = 'mac64'



import os
import requests
import zipfile
import platform as pl
import xml.etree.ElementTree as ElTree
import subprocess
import struct
from urllib.request import urlopen
from bs4 import BeautifulSoup

from requests.exceptions import RequestException
from requests.exceptions import HTTPError

from . import downloader, constants
from .enums import Platform, Phase
from .exceptions import GetChromeDriverError, UnknownPlatformError, VersionUrlError, UnknownVersionError, \
    DownloadError, VersionError


class GetChromeDriver:

    def __init__(self, platform=None):
        if not platform:
            if pl.system() == 'Windows':
                self.__platform = self.__check_platform(Platform.win)
            elif pl.system() == 'Linux':
                self.__platform = self.__check_platform(Platform.linux)
            elif pl.system() == 'Darwin':
                self.__platform = self.__check_platform(Platform.mac)
        else:
            self.__platform = self.__check_platform(platform)

    def stable_version(self) -> str:
        """ Return the latest stable version """

        return self.__latest_version_by_phase(Phase.stable)

    def beta_version(self) -> str:
        """ Return the latest beta version """

        return self.__latest_version_by_phase(Phase.beta)

    def __latest_version_by_phase(self, phase) -> str:
        """
        Return the latest stable or latest beta version

        :param phase: Stable or beta
        """

        response = requests.get(constants.url_chromium)
        if not response.ok:
            raise GetChromeDriverError(f'Could not get {constants.url_chromium}')

        soup = BeautifulSoup(response.content, 'html.parser')
        ul = soup.select_one(constants.css_selector_versions)
        for li in ul:
            li_text = li.text.replace(u'\u00A0', ' ')
            if Phase.stable == phase:
                version_str = constants.latest_stable_version_str
            else:
                version_str = constants.latest_beta_version_str

            if li_text[:len(version_str)].lower() == version_str.lower():
                try:
                    version = li.a['href'].split('path=')[-1:][0][:-1]
                except TypeError:
                    raise UnknownVersionError('Could not find version')
                else:
                    self.__check_if_version_format_is_valid(version)
                    return version
        else:
            raise UnknownVersionError('Could not find version')

    def stable_version_url(self) -> str:
        """ Return the latest stable version url """

        return self.version_url(self.__latest_version_by_phase(Phase.stable))

    def beta_version_url(self) -> str:
        """ Return the latest beta version url """

        return self.version_url(self.__latest_version_by_phase(Phase.beta))

    def version_url(self, version: str) -> str:
        """
        Return the version download url

        :param version: Chromedriver version
        """

        self.__check_if_version_format_is_valid(version)
        arch = struct.calcsize('P') * 8
        arch_64 = 64
        chromedriver = 'chromedriver'
        zip_ext = '.zip'

        if self.__platform == Platform.win:
            # 64bit
            if arch == arch_64:
                try:
                    url = f'{constants.url_chromedriver_storage}/{version}/{chromedriver}_{Platform.win64}{zip_ext}'
                    self.__check_if_url_is_valid(url)
                    return url
                except VersionUrlError:
                    # No 64 bit, get 32 bit
                    pass
            # 32bit
            url = f'{constants.url_chromedriver_storage}/{version}/{chromedriver}_{Platform.win32}{zip_ext}'
            self.__check_if_url_is_valid(url)
            return url

        elif self.__platform == Platform.linux:
            # 64bit
            if arch == arch_64:
                try:
                    url = f'{constants.url_chromedriver_storage}/{version}/{chromedriver}_{Platform.linux64}{zip_ext}'
                    self.__check_if_url_is_valid(url)
                    return url
                except VersionUrlError:
                    # No 64 bit, get 32 bit
                    pass
            # 32bit
            url = f'{constants.url_chromedriver_storage}/{version}/{chromedriver}_{Platform.linux32}{zip_ext}'
            self.__check_if_url_is_valid(url)
            return url

        elif self.__platform == Platform.mac:
            # 64bit
            if arch == arch_64:
                try:
                    url = f'{constants.url_chromedriver_storage}/{version}/{chromedriver}_{Platform.mac64}{zip_ext}'
                    self.__check_if_url_is_valid(url)
                    return url
                except VersionUrlError:
                    # No 64 bit, get 32 bit
                    pass
            # 32bit
            url = f'{constants.url_chromedriver_storage}/{version}/{chromedriver}_{Platform.mac32}{zip_ext}'
            self.__check_if_url_is_valid(url)
            return url

    def download_stable_version(self, output_path=None, extract=False):
        """
        Download the latest stable chromedriver version

        :param output_path: Path to download the driver to
        :param extract: Extract the downloaded driver or not
        """

        version = self.__latest_version_by_phase(Phase.stable)
        self.download_version(version=version, output_path=output_path, extract=extract)

    def download_beta_version(self, output_path=None, extract=False):
        """
        Download the latest beta chromedriver version

        :param output_path: Path to download the driver to
        :param extract: Extract the downloaded driver or not
        """

        version = self.__latest_version_by_phase(Phase.beta)
        self.download_version(version=version, output_path=output_path, extract=extract)

    def download_version(self, version, output_path=None, extract=False) -> str:
        """
        Download a chromedriver version

        :param version: Chromedriver version
        :param output_path: Path to download the driver to
        :param extract: Extract the downloaded driver or not
        """

        self.__check_if_version_format_is_valid(version)
        if not output_path:
            # on path is None,
            # ChromeDriver will be downloaded at e.g. chromedriver/88.0.4324.96/bin/chromedriver.exe
            output_path = self._output_path(version)

        # If path == webdriver/bin (or any other dir name)
        # ChromeDriver will be downloaded at webdriver/bin/chromedriver.exe

        def download(download_url, download_output_path) -> str:
            try:
                file_path, file_name = downloader.download(url=download_url, output_path=download_output_path)
            except (OSError, HTTPError, RequestException) as err:
                raise DownloadError(err)
            if extract:
                with zipfile.ZipFile(file_path, 'r') as zip_ref:
                    zip_ref.extractall(path=download_output_path)
                os.remove(file_path)
                if self.__platform == Platform.linux or self.__platform == Platform.mac:
                    os.chmod(f'{download_output_path}/chromedriver', 0o755)
            return download_output_path

        # Download the driver file and return the path of the driver file
        url = self.version_url(version)
        return download(download_url=url, download_output_path=output_path)

    def __check_if_url_is_valid(self, url) -> None:
        """
        Check if url is valid

        :param url: The driver download url
        """

        if requests.head(url).status_code != 200:
            raise VersionUrlError('Invalid url')

    def __check_if_version_format_is_valid(self, version):
        """
        Check if version format is valid

        :param version: Chromedriver version
        """

        split_version = version.split('.')
        for number in split_version:
            if not number.isnumeric():
                raise UnknownVersionError('Invalid version format')

    def __check_platform(self, platform) -> str:
        """
        Check if platform is valid

        :param platform: OS
        """

        platforms_list = [platform for platform in Platform]

        if platform not in platforms_list:
            raise UnknownPlatformError(f'Unknown platform, choose a platform from: {str(platforms_list)}')
        return platform

    def matching_version(self) -> str:
        """ Return a matching ChromeDriver version """

        all_chromedriver_versions = self.__get_all_chromedriver_versions()
        installed_chrome_version = self.__get_installed_chrome_version()
        for chromedriver_version in reversed(all_chromedriver_versions):
            if '.'.join(installed_chrome_version.split('.')[:-1]) == '.'.join(chromedriver_version.split('.')[:-1]):
                return chromedriver_version

    def auto_download(self, output_path=None, extract=False) -> str:
        """
        Download ChromeDriver for the installed Chrome version on machine

        :param output_path: Path to download the driver to
        :param extract: Extract the downloaded driver or not
        """

        version = self.matching_version()
        if version == '' or version is None:
            raise VersionError('Unable to find a ChromeDriver version for the installed Chrome version')
        return self.download_version(version, output_path, extract)

    def install(self):
        """ Install ChromeDriver for the installed Chrome version on machine """

        output_path = self.auto_download(extract=True)
        path = os.path.join(os.path.abspath(os.getcwd()), output_path)
        os.environ['PATH'] += os.pathsep + os.pathsep.join([path])

    def __get_all_chromedriver_versions(self) -> list:
        """ Return a list with all ChromeDriver versions """

        key_texts = []
        versions = []

        with urlopen(constants.url_chromedriver_storage) as xml_file:
            tree = ElTree.parse(xml_file)
            root = tree.getroot()

            for root_item in root:
                # Remove namespace
                root_item.tag = root_item.tag.split('}', 1)[1]

            for content in root.findall('Contents'):

                for content_item in content:
                    # Remove namespace
                    content_item.tag = content_item.tag.split('}', 1)[1]

                    key_texts.append(content.find('Key').text)

        for text in key_texts:

            version = ''
            for char in text:
                if char.isnumeric() or char == '.':
                    version += char
                else:
                    break

            if len(version) < 1:
                continue

            versions.append(version)

        return list(dict.fromkeys(versions))

    def __get_installed_chrome_version(self) -> str:
        """ Return the installed Chrome version on the machine """

        if self.__platform == Platform.win:
            process = subprocess.Popen(
                ['reg', 'query', 'HKEY_CURRENT_USER\\SOFTWARE\\Google\\Chrome\\BLBeacon', '/v', 'version'],
                stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, stdin=subprocess.DEVNULL)
            return process.communicate()[0].decode('UTF-8').split()[-1]

        elif self.__platform == Platform.linux:
            process = subprocess.Popen(
                ['google-chrome', '--version'],
                stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, stdin=subprocess.DEVNULL)
            return process.communicate()[0].decode('UTF-8').split()[-1]

        elif self.__platform == Platform.mac:
            process = subprocess.Popen(
                ['/Applications/Google Chrome.app/Contents/MacOS/Google Chrome', '--version'],
                stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, stdin=subprocess.DEVNULL)
            return process.communicate()[0].decode('UTF-8').split()[-1]

    def _output_path(self, version) -> str:
        """
        Return the output path

        :param version: Chromedriver version
        """

        return f'chromedriver/{version}/bin'



url_chromium = 'https://chromedriver.chromium.org'
url_chromedriver_storage = 'https://chromedriver.storage.googleapis.com'
css_selector_versions = 'ul.n8H08c:nth-child(5)'
latest_stable_version_str = 'Latest stable release'
latest_beta_version_str = 'Latest beta release'



import typer

from . import __version__
from .get_driver import GetChromeDriver
from .enums import Phase, Platform
from .exceptions import GetChromeDriverError

app = typer.Typer(name='Get ChromeDriver', add_completion=False)


@app.command()
def main(beta_version: bool = typer.Option(default=False,
                                           help='Print the beta version',
                                           show_default=False),

         stable_version: bool = typer.Option(default=False,
                                             help='Print the stable version',
                                             show_default=False),

         latest_urls: bool = typer.Option(default=False,
                                          help='print the beta and stable version download urls for all platforms',
                                          show_default=False),

         version_url: str = typer.Option(default=None,
                                         help='Print the version download url',
                                         show_default=False),

         beta_url: bool = typer.Option(default=False,
                                       help='Print the beta version download url',
                                       show_default=False),

         stable_url: bool = typer.Option(default=False,
                                         help='Print the stable version download url',
                                         show_default=False),

         auto_download: bool = typer.Option(default=False,
                                            help='Download a ChromeDriver version for the installed Chrome Version',
                                            show_default=False),

         download_beta: bool = typer.Option(default=False,
                                            help='Download beta version',
                                            show_default=False),

         download_stable: bool = typer.Option(default=False,
                                              help='Download stable version',
                                              show_default=False),

         download_version: str = typer.Option(default=None,
                                              help='Download a specific version',
                                              show_default=False),

         extract: bool = typer.Option(default=False,
                                      help='Extract the compressed driver file',
                                      show_default=False),

         version: bool = typer.Option(default=False,
                                      help='Application version',
                                      show_default=False)):
    """ Main """

    if beta_version:
        __print_latest_version(phase=Phase.beta)

    elif stable_version:
        __print_latest_version(phase=Phase.stable)

    elif latest_urls:
        __print_latest_urls()

    elif version_url:
        __print_version_url(version=version_url)

    elif beta_url:
        __print_latest_url(phase=Phase.beta)

    elif stable_url:
        __print_latest_url(phase=Phase.stable)

    elif auto_download:
        __auto_download(extract=extract)

    elif download_beta:
        __download_latest_version(phase=Phase.beta, extract=extract)

    elif download_stable:
        __download_latest_version(phase=Phase.stable, extract=extract)

    elif download_version:
        __download_version(version=download_version, extract=extract)

    elif version:
        print(f'v{__version__}')


def __print_latest_version(phase: Phase):
    """
    Print latest stable version or latest beta version

    :param phase: Stable or beta
    """

    get_driver = GetChromeDriver()
    error = 'No latest version found'
    if phase == Phase.beta:
        try:
            print(get_driver.beta_version())
        except GetChromeDriverError:
            print(error)
    elif phase == Phase.stable:
        try:
            print(get_driver.stable_version())
        except GetChromeDriverError:
            print(error)
    else:
        print(error)


def __print_latest_urls():
    """ Print the stable and beta url version for all platforms """

    get_driver_win = GetChromeDriver(Platform.win)
    get_driver_linux = GetChromeDriver(Platform.linux)
    get_driver_mac = GetChromeDriver(Platform.mac)
    get_drivers = {'Windows': get_driver_win, 'Linux': get_driver_linux, 'macOS': get_driver_mac}

    error = 'Could not get latest urls'

    result = ''
    for index, (key, value) in enumerate(get_drivers.items()):
        try:
            result += f'Latest beta and stable version for {key}:\n'
            result += f'stable : {value.stable_version_url()}\n'
            result += f'beta   : {value.beta_version_url()}'
            if index < len(get_drivers) - 1:
                result += '\n'
        except GetChromeDriverError:
            print(error)

    print(result)


def __print_version_url(version: str):
    """
    Print the url of a version

    :param version: Chromedriver version
    """

    get_driver = GetChromeDriver()

    error = 'Could not find version url'

    try:
        print(get_driver.version_url(version))
    except GetChromeDriverError:
        print(error)


def __print_latest_url(phase: Phase):
    """
    Print latest stable url or latest beta url

    :param phase: Stable or beta
    """

    get_driver = GetChromeDriver()

    error = 'Could not find version url'

    if phase == Phase.beta:
        try:
            print(get_driver.beta_version_url())
        except GetChromeDriverError:
            print(error)
    elif phase == Phase.stable:
        try:
            print(get_driver.stable_version_url())
        except GetChromeDriverError:
            print(error)


def __auto_download(extract: bool):
    """
    Auto download driver

    :param extract: Extract the downloaded driver or not
    """

    get_driver = GetChromeDriver()

    try:
        get_driver.auto_download(extract=extract)
        print('Download finished')
    except GetChromeDriverError:
        print('An error occurred at downloading')


def __download_latest_version(phase: Phase, extract: bool = False):
    """
    Download the driver for the stable version or beta version

    :param phase: Stable or beta
    :param extract: Extract the downloaded driver or not
    """

    get_driver = GetChromeDriver()

    download_complete = 'Download complete'
    beta_error = 'Could not download beta version'
    stable_error = 'Could not download stable version'

    if phase == Phase.beta:
        try:
            get_driver.download_beta_version(extract=extract)
            print(download_complete)
        except GetChromeDriverError:
            print(beta_error)
    elif phase == Phase.stable:
        try:
            get_driver.download_stable_version(extract=extract)
            print(download_complete)
        except GetChromeDriverError:
            print(stable_error)


def __download_version(version: str, extract: bool):
    """
    Download driver version

    :param version: Chromedriver version
    :param extract: Extract the downloaded driver or not
    """

    get_driver = GetChromeDriver()

    try:
        get_driver.download_version(version=version, extract=extract)
        print('Download finished')
    except GetChromeDriverError:
        print('Could not download version')
